//==============================================================
// Copyright  2020 Intel Corporation
//
// SPDX-License-Identifier: MIT
// =============================================================

#pragma once

#ifndef MATRIX_MUL_RPI4B_H
#define MATRIX_MUL_RPI4B_H

constexpr std::size_t N = 10; // # of rows
constexpr std::size_t M = 20; // # of cols

// Using Mersenne Twister (19937) algoritm for generating values
std::mt19937 gen((unsigned)std::time(nullptr));

// Declaring the default CL/SYCL kernels queue (ordered execution)
// Note: Queues with GPGPUs/FPGAs non-default device selector will 
// simply *NOT* work on Raspberry PI 4B+ and other IoT boards
cl::sycl::queue q = cl::sycl::queue{};

// alloc_matrix(...) routine allocates a memory buffer 
// for matrix A, consisting of NxM elements of type short
template<class _Alloc>
void alloc_matrix(short**& A, \
	std::size_t N, std::size_t M, _Alloc allocator)
{
	if (A == nullptr) {
		// Allocating memory for N-buffers using std::malloc(...)
		A = static_cast<short**>(std::malloc(sizeof(short*) * N));

		// Allocating memory for each ii-th buffer of size M
		for (std::size_t ii = 0; ii < N; ii++)
		{
			// Allocating the ii-th row's memory buffer explicitly
			A[ii] = static_cast<short*>(\
				std::malloc(sizeof(short) * M));

			if (A[ii] != nullptr) {
				// If the allocation succeeded, invoking the allocator lambda-function
				// to initialize each element in the matrix A with a zero or random value
				allocator(A[ii], M);
			}
		}
	}
}

// rand_matrix(...) is a specialization of the alloc_matrix(...) routine,
// used for allocating matrix A memory buffers, generating a random value
// for each of its elements using the C++17's std::generate STL-algorithm
void rand_matrix(short**& A, \
	std::size_t N, std::size_t M) {
	// Instantiate the STL's random uniform distribution object
	std::uniform_int_distribution<> uid(0, 9);
	// Do the matrix A's memory buffers allocation by calling the
	// alloc_matrix(...) routine and generating random values for
	// the elements of each of N-rows in the matrix A, by using the 
	// STL's random numbers generator routine uid(gen).
	alloc_matrix(A, N, M, \
		[&](short*& r, const std::size_t& count) {
			std::generate(r, r + count, [&]() { return uid(gen); });
		});
}

// mat_mul_sycl(...) is a routine that instantiates a CL/SYCL kernel,
// implementing the actual matrix multiplication, and, finally, submiting 
// the kernel's code to the global CL/SYCL-queue for the execution
void mat_mul_sycl(short**& A, short**& B, short**& C)
{
	// Do the resultant matrix C's memory buffer allocation, assigning
	// an initial value of "zero" to each element of the matrix C
	alloc_matrix(C, N, N, \
		[&](short*& r, const std::size_t& count) {
			std::fill(r, r + count, 0L);
		});

	// Submitting the matrix multiplication kernel to the
	// global CL/SYCL-queue by calling q.submit(...) method
	q.submit([&](cl::sycl::handler& cgh) {
		// Invoking the CL/SYCL-queue's parallel_for tight loop algorithm,
		// collapsing the two nested loops into a single loop, each iteration
		// of which is executed in parallel, performing the scalar multiplication
		// of ii-th and jj-th vectors of matrices A and B, respectively
		cgh.parallel_for(cl::sycl::range<2>{N, N}, \
			[=, a_ptr = A, b_ptr = B, c_ptr = C](cl::sycl::id<2> idx) {
				// Obtaining the index values of matrix A's ii-th 
				// row and matrix B's jj-th column
				std::size_t ii = static_cast<std::size_t>(idx[0]);
				std::size_t jj = static_cast<std::size_t>(idx[1]);

				// Performing the scalar multiplication of the A's ii-th row and
				// B's jj-th column vectors, respectively, accumulating the result
				// in the private variable scalar
				short scalar = 0;
				for (int kk = 0; kk < M; kk++) {
					scalar += a_ptr[ii][kk] * b_ptr[kk][jj];
				}

				// If the scalar product value is not equal to 0, then assigning it to
				// an element of the resultant matrix C with indices of ii and jj, respectively
				if (scalar > 0) {
					c_ptr[ii][jj] = scalar;
				}
			});
		});

	// Barrier synchronization
	q.wait();
}

void print_matrix(short** A, std::size_t N, std::size_t M)
{
	// Performing an output of the matrix A's elements
	for (std::size_t ii = 0; ii < N; ii++)
	{
		for (std::size_t jj = 0; jj < M; jj++)
			std::cout << A[ii][jj] << " ";

		std::cout << "\n";
	}

	std::cout << "\n";
}

#endif // MATRIX_MUL_RPI4B_H